#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "debug.h"
#include "complex.h"

#define min(a, b) (((a) < (b)) ? (a) : (b))

#define SHARP (-1)

#define Unit (Val_int (0))

#define REPULSE 3.0
#define STABLE_LENGTH 1.0
#define ATTRACT 5.0
#define VISCOSITY 0.5
#define RUBBING 0.0

#define DT 0.05

#define BORED 10000
#define EPSILON 0.05

#define BORDER 0

static value G; /* le graphe a dessiner */
static int n, d; /* la taille du graphe */

static cplx_t *pos, *speed, *accel;
static double x_min, x_max, y_min, y_max;
static double v_max;

static int width, height;

/* le k-ieme voisin de node */
int neigbhoor (int node, int k)
{  
  return Int_val (Field (Field (G, node), k)); 
}
void print (cplx_t *tab)
{
  int i;
  return;

  for (i = 0; i < n; i++) {
    printf ("%d : (%g, %g)\n", i, tab[i].re, tab[i].im);
  }
}

int *rand_tab (int n)
{
  int i, *t;
  debug_in (0, "Calculating random permutation...");
  t = malloc (n * sizeof (int));
  assert (t != NULL);

  t[0] = 0;
  for (i = 1; i < n; i++) {
    int p = rand () % (i + 1);
    t[i] = t[p];
    t[p] = i;
  }
  debug_out (0, "Done.");
  return t;
}

double node_repulse (double dist)
{  return exp (-REPULSE * dist); }

double edge_attract (double dist)
{  return (dist - STABLE_LENGTH) * ATTRACT; }

double rubbing (double speed)
{ return speed * VISCOSITY; }


void init_pos (void)
{
  int k, *t;
  debug_in (0, "Initializing positions...");
  t = rand_tab (n);
  for (k = 0; k < n; k++) {
    ni_root (pos + k, n, t[k]);
    //    cr_mul (pos + k, (double) n / 2);
    accel[k].re = accel[k].im = 0.0;
    speed[k].re = speed[k].im = 0.0;
  }
  free (t);
  debug_out (0, "Done.");
}



value compute_int_pos (void)
{
  int i;
  double ratio = min (((double) (width - 2 * BORDER)) / (x_max - x_min),
		      ((double) (height - 2 * BORDER)) / (y_max - y_min));
/*    double ax = (((double) width) - ratio * (x_max - x_min)) / 2.0; */
/*    double ay = (((double) height) - ratio * (y_max - y_min)) / 2.0; */
  CAMLparam0 ();
  CAMLlocal1 (ipos);

  debug_in (0, "Computing integer positions...");

  ipos = alloc_tuple (n);

  for (i = 0; i < n; i++) {
    Store_field (ipos, i, alloc_tuple (2));
    Store_field (Field (ipos, i),
		 0,
		 Val_int (width / 2 + (int) (ratio * (pos[i].re - (x_min + x_max) / 2.0))));
    Store_field (Field (ipos, i),
		 1,
		 Val_int (width / 2 + (int) (ratio * (pos[i].im - (y_min + y_max) / 2.0))));
  }
  debug_out (0, "Done.");
  CAMLreturn (ipos);
}


void alloc_all  (value graph, value w, value h) 
{
  CAMLparam3 (graph, w, h);
  debug_in (0, "Allocating...");

  register_global_root (&G);
  G = graph;
  n = Wosize_val (graph);
  d = Wosize_val (Field (graph, 0));

  width = Int_val (w);
  height = Int_val (h);

  pos = malloc (sizeof (cplx_t) * n);
  assert (pos != NULL);

  speed = malloc (sizeof (cplx_t) * n);
  assert (speed != NULL);

  accel = malloc (sizeof (cplx_t) * n);
  assert (accel != NULL);

  debug_out (0, "Done.");
  CAMLreturn0;
}

void free_all (void)
{
  debug_in (0, "Freeing global vars...");
  remove_global_root (&G);

  free (pos); 
  free (speed); 
  free (accel);
  debug_out (0, "Done.");
}

void compute_pos (void)
{
  cplx_t tmp;
  int i;
  debug_in (1, "Computing positions...");
  for (i = 0; i < n; i++) {
    /* pos <- pos + speed * dt */
    tmp = speed[i];
    cr_mul (&tmp, DT);
    cc_add (pos + i, &tmp);
    if (pos[i].re < x_min) x_min = pos[i].re;
    if (pos[i].re > x_max) x_max = pos[i].re;
    if (pos[i].im < y_min) y_min = pos[i].im;
    if (pos[i].im > y_max) y_max = pos[i].im;
  }
  debug_out (1, "Done.");
}

void compute_speed (void)
{
  cplx_t tmp;
  double v; 
  double ratio = min (((double) (width - 2 * BORDER)) / (x_max - x_min),
		      ((double) (height - 2 * BORDER)) / (y_max - y_min));
  int i;
  debug_in (1, "Computing speeds...");

  v_max = 0.0;
  for (i = 0; i < n; i++) {
    /* speed <- speed + accel * dt */
    tmp = accel[i];
    cr_mul (&tmp, DT);
    cc_add (speed + i, &tmp);
    if ((v = c_norm (speed + i)) > v_max)
      v_max = v;
  }
  v_max = v_max * DT * ratio;
  debug_out (1, "Done.");
}



void compute_accel (void)
{
  int i, j, v;
  cplx_t F;
  double dist;
  debug_in (1, "Computing accelerations...");

  for (i = 0; i < n; i++) {
    accel[i].re = accel[i].im = 0.0;
  }

  /* compute node repulses for each _other_ node*/
  debug_in (1, "Computing repulses...");
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++) {
      if (i == j) continue;
      F = pos[i];
      cc_sub (&F, pos + j);
      dist = c_norm (&F);
      if (dist == 0.0) continue;
      cr_mul (&F, node_repulse (dist) / dist);
      cc_add (accel + i, &F);
    }
  debug_out (1, "Done");
  /* compute edge attracts */
  debug_in (1, "Computing attracts...");
  for (i = 0; i < n; i++)
    for (j = 0; j < d; j++) {
      if ((v = neigbhoor (i, j)) == SHARP) continue;
      F = pos[v];
      cc_sub (&F, pos + i);
      dist = c_norm (&F);
      if (dist == 0.0) continue;
      cr_mul (&F, edge_attract (dist) / dist);
      cc_add (accel + i, &F);
    }
  debug_out (1, "Done");

  /* compute rubbings */
  debug_in (1, "Computing rubbings...");
  for (i = 0; i < n; i++) {
    cplx_t V = speed[i];
    F = V;
    dist = c_norm (&F);
    if (dist == 0.0) continue;
    if (dist < RUBBING) {
      accel[i].re = accel[i].im = 0.0;
      continue;
    }
    cr_mul (&V, - RUBBING / dist);
    cr_mul (&F, rubbing (dist) / dist);
    cc_sub (accel + i, &F);
    cc_add (accel + i, &V);
  }
  debug_out (1, "Done");
  debug_out (1, "Done.");
}

value caml_positions (value graph, value w, value h)
{
  int count = 0;
  //  static int num_appel = 0;
  CAMLparam3 (graph, w, h);
  CAMLlocal1 (ipos);

/*    if (num_appel++ >= 39) */
/*      set_debug_level (0); */
/*    else */
    set_debug_level (-2);
  debug_in (-1, "Enter C section...");

  alloc_all (graph, w, h);
  init_pos ();
  
  do {
    compute_accel ();
    compute_speed ();
    compute_pos ();
    if (count++ > BORED) {
      fprintf (stderr, "tired of computing...\n");
      break;
    }
  } while (v_max > EPSILON);

  free_all ();
  ipos = compute_int_pos ();

  debug_out (-1, "Done.");
  CAMLreturn (ipos);
}





